from .api import Estado  # , Municipio, OrgaoExpedidor
from .settings import UF_MAP, NUM_MAP
from .errors import EstadoNotFoundError


class Extrator:
    def __init__(self, cookie):
        self._headers = {
            'authority': 'portalbnmp.cnj.jus.br',
            'accept': 'application/json',
            'user-agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_8_2) AppleWebKit/537.17 (KHTML, like Gecko)\
     Chrome/24.0.1309.0 Safari/537.17',
            'content-type': 'application/json;charset=UTF-8',
            'origin': 'https://portalbnmp.cnj.jus.br',
            'sec-fetch-site': 'same-origin',
            'sec-fetch-mode': 'cors',
            'sec-fetch-dest': 'empty',
            'referer': 'https://portalbnmp.cnj.jus.br/',
            'accept-language': 'en-US,en;q=0.9',
            'cookie': cookie,
        }

    def estado(self, sigla: [int, str]):
        if isinstance(sigla, str):
            sigla = sigla.upper()
            if sigla in UF_MAP:
                return Estado(self._headers, UF_MAP[sigla])
        elif isinstance(sigla, int):
            if sigla in NUM_MAP:
                return Estado(self._headers, sigla)
        raise EstadoNotFoundError("id_estado precisa ser a sigla de uma UF ou um número entre 1 e 27")



from setuptools import setup
from pathlib import Path


setup(
    name='bnmp_scraper',
    version='0.0.1',

    url='https://github.com/olucaslopes/BNMP-Scraper',
    author='Lucas Lopes',
    author_email='lucaslopesamorim@gmail.com',
    description='Navega pela API do Portal BNMP coletando dados',
    long_description=Path("README.md").read_text(encoding="utf-8"),
    long_description_content_type="text/markdown",

    packages=['bnmp_scraper'],
    install_requires=[
        'requests',
        'tqdm'
    ],
)



# from .api import Estado, Municipio, OrgaoExpedidor
from .errors import MandadosNotFoundError
import requests
import concurrent.futures
from itertools import repeat
from .settings import PARAMS_FORCA_BRUTA
from tqdm import tqdm


class Filtro:
    def __init__(self, headers):
        self._mandados_list = []
        self._headers = headers

    def obter_mandados(self) -> list:
        if not self._mandados_list:
            raise MandadosNotFoundError("Você ainda não baixou os mandados. Utilize .baixar_mandados() para baixa-los")
        return self._mandados_list.copy()

    # def __add__(self, val):
    #     if not isinstance(val, (Estado, Municipio, OrgaoExpedidor)):
    #         raise TypeError("Você só pode somar elementos das classes Estado, Municipio ou OrgaoExpedidor")
    #     return self.obter_mandados() + val.obter_mandados()

    def _request_post(self, params, data):
        """Faz uma requisição do tipo POST e retorna
        um JSON se ela for bem sucedida e um NoneType
        caso contrário"""
        response = requests.post(
            url='https://portalbnmp.cnj.jus.br/bnmpportal/api/pesquisa-pecas/filter',
            headers=self._headers,
            params=params,
            data=data
        )
        if response.ok:
            return response.json()
        else:
            print(f"Deu Ruim! POST request não bem sucedida. Status code:{response.status_code}")

    def _obter_data_post(self, id_estado: int, id_municipio: int = "", id_orgao: int = "") -> str:
        """Tem como parâmetros os ids em questão.
        Retorna a variável data necessário para
        fazer uma requisição do tipo POST."""
        if not id_municipio and not id_orgao:
            # Só tem estado!
            return '{"buscaOrgaoRecursivo":false,"orgaoExpeditor":{},"idEstado":' + str(id_estado) + '}'
        elif not id_orgao:
            # Tem estado e id_municipio!
            return '{"buscaOrgaoRecursivo":false,"orgaoExpeditor":{},"idEstado":' + str(
                id_estado) + ',"idMunicipio":' + str(id_municipio) + '}'
        else:
            # Tem estado, municipio e orgao!
            return '{"buscaOrgaoRecursivo":false,"orgaoExpeditor":{"id":' + str(id_orgao) + '},"idEstado":' + str(
                id_estado) + ',"idMunicipio":' + str(id_municipio) + '}'

    def _obter_post_pag1(self, id_estado: int, id_municipio: int = 0, id_orgao: int = 0, size: int = 10) -> dict:
        """Faz um POST request da primeira página para
        obter até 2.000 elementos. Retorna um dicionário
         e um int com o total de mandados daquele id."""
        params = (
            ('page', '0'),
            ('size', f'{size}'),
            ('sort', 'dataExpedicao,ASC'),
        )

        data = self._obter_data_post(id_estado, id_municipio, id_orgao)

        raw_data = self._request_post(params, data)
        return raw_data

    def _request_post_poucos_mandados(self, response_pag1: dict, id_estado: int, id_municipio: int = 0,
                                      id_orgao: int = 0) -> list:
        """
        Obtém até 10.000 linhas de dados.
        Retorna uma lista de dicts
        :param response_pag1: 'raw_data': dict
        :param id_estado: id que pertence ao intervalo [1,27].
        :param id_municipio: id interno
        :param id_orgao: id interno
        :return: list: lista de dicionários de mandados
        """
        data = self._obter_data_post(id_estado, id_municipio, id_orgao)
        tot_elements = response_pag1.get('totalElements')
        tot_pages = tot_elements // 2000 if tot_elements % 2000 == 0 else tot_elements // 2000 + 1
        params = tuple((('page', str(x)), ('size', '2000'), ('sort', 'dataExpedicao,ASC')) for x in range(0, tot_pages))
        with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
            novos_mandados = list(executor.map(self._request_post, params, repeat(data)))
        return [x for sublist in novos_mandados for x in sublist['content']]  # List unpacking

    def _request_post_expandido(self, response_pag1: dict, id_estado: int, id_municipio: int = 0,
                                id_orgao: int = 0) -> list:
        """
        Aproveita-se da ordenação para extrapolar o
        limite de 10.000 linhas e alcaçar até o dobro disso.
        Retorna uma lista de dicts
        """
        data = self._obter_data_post(id_estado, id_municipio, id_orgao)
        tot_elements = response_pag1['totalElements']
        params = tuple((('page', str(x)), ('size', '2000'), ('sort', 'dataExpedicao,ASC')) for x in range(0, 5))
        with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
            novos_mandados = list(executor.map(self._request_post, params, repeat(data)))
        all_mandados = [x for sublist in novos_mandados for x in sublist['content']]  # List unpacking
        paginas_restantes = (tot_elements - 10000) // 2000
        if paginas_restantes >= 1:
            params = tuple(
                (('page', str(x)), ('size', '2000'),
                 ('sort', 'dataExpedicao,DESC')) for x in range(0, int(paginas_restantes)))
            with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
                novos_mandados = executor.map(self._request_post, params, repeat(data))
            for e in [x['content'] for x in novos_mandados]:
                all_mandados.extend(e)
        params = (('page', str(paginas_restantes)), ('size', '2000'), ('sort', 'dataExpedicao,DESC'))
        ultima_pag = self._request_post(params, data)
        all_mandados.extend([x for x in ultima_pag['content'] if x not in all_mandados])
        return all_mandados

    def _request_post_forcabruta(self, id_estado: int, id_municipio: int, id_orgao: int) -> list:
        """
        Aplica todas as ordenações disponíveis para
        obter o maior número de mandados possíveis,
        mesmo sem conseguir acessar todas as páginas
        do Banco de Dados.
        Retorna uma lista com todos os mandados adquiridos.
        """
        data = '{"buscaOrgaoRecursivo":false,"orgaoExpeditor":{"id":' + str(id_orgao) + '},"idEstado":' + str(
            id_estado) + ',"idMunicipio":' + str(id_municipio) + '}'

        all_mandados = list()

        with concurrent.futures.ThreadPoolExecutor(max_workers=50) as executor:
            novos_mandados = list(
                tqdm(executor.map(self._request_post, PARAMS_FORCA_BRUTA, repeat(data)), total=len(PARAMS_FORCA_BRUTA)))
        for lista_mandados in [x['content'] for x in novos_mandados if x]:
            all_mandados.extend([mandado for mandado in lista_mandados if mandado not in all_mandados])
        return all_mandados

    def _pega_conteudo_completo(self, linha: dict):
        """
        "Entra" em cada um dos mandados com o
        método GET para pegar mais informações
        e salva-nas em json/<id_mandado>.json
        """
        if not isinstance(linha, dict):
            print(f"ERROR: Mandado is {type(linha)}")
            return None
        # HEADERS['user-agent'] = ua.random
        id_mandado = linha.get("id")
        id_tipo_peca = linha.get("idTipoPeca")
        if id_tipo_peca is None:
            id_tipo_peca = linha['tipoPeca'].get('id')

        response = requests.get(
            url=f'https://portalbnmp.cnj.jus.br/bnmpportal/api/certidaos/{id_mandado}/{id_tipo_peca}',
            headers=self._headers
        )
        if response.ok:
            return response.json()
            # with open(f"jsons/{response_dict['id']}.json", 'wb') as outf:
            #     outf.write(response.content)
        else:
            print(f"Deu ruim! Status code: {response.status_code}")
            # print("Você está no pega_conteudo_completo")
            # pass

    def _baixar_conteudo_completo_parallel(self, lista_mandados: list) -> list:
        """
        Dada uma lista de mandados, utiliza concorrência
        para salvar cada mandado em um arquivo JSON de
        maneira mais eficiente.
        """
        executor = concurrent.futures.ThreadPoolExecutor(max_workers=32)
        result = list(executor.map(self._pega_conteudo_completo, lista_mandados))
        return [mandado for mandado in result if mandado is not None]



import requests
from .settings import *
# from .utils import obter_post_pag1, request_post_poucos_mandados, request_post_expandido, \
#     request_post_forcabruta, baixar_conteudo_completo_parallel
from .errors import MandadosNotFoundError, EstadoNotFoundError
import concurrent.futures
from tqdm import tqdm
from .filtro import Filtro
# from extrator import Extrator


class Estado(Filtro):
    def __init__(self, headers, id_estado: [int, str]):
        super().__init__(headers)
        self._id = self._set_id(id_estado)
        self._init_info = self._obter_post_pag1(self._id)
        self._totalElements = self._init_info['totalElements']
        self._munic_info = {}
        self._munic_objs = []

    def baixar_mandados(self, force: bool = False):
        if self._mandados_list and not force:
            return self._mandados_list
        if self._totalElements <= 20000:
            if self._totalElements <= 10000:
                mandados_parciais = self._request_post_poucos_mandados(self._init_info, self._id)
            else:
                mandados_parciais = self._request_post_expandido(self._init_info, self._id)
            self._mandados_list = self._baixar_conteudo_completo_parallel(mandados_parciais)
        else:
            municips = self._gerar_municipios()
            executor = concurrent.futures.ThreadPoolExecutor(max_workers=32)
            list(tqdm(executor.map(Municipio.baixar_mandados, municips), total=len(municips)))
            self._mandados_list = [item for munic in municips for item in munic.obter_mandados()]
        print(f"Recuperados {len(self._mandados_list)}/{self._totalElements} mandados")

    def obter_municicipios(self) -> list:
        """
        Retorna uma lista de ids dos municipios dessa uf.
        """
        return list(self._baixar_municipios().keys())

    def _baixar_municipios(self, force: bool = False) -> dict:
        """
        A partir do id da de uma UF retorna um
        dict mapeando todos os municípios dessa
        UF aos seus ids.
        """
        if self._munic_info and not force:
            return self._munic_info
        response_munic = requests.get(
            url=f'https://portalbnmp.cnj.jus.br/scaservice/api/municipios/por-uf/{self._id}',
            headers=HEADERS
        )
        munic_list = response_munic.json()
        self._munic_info = {munic['nome']: munic['id'] for munic in munic_list}
        return self._munic_info

    def _gerar_municipios(self, force: bool = False) -> list:
        """"
        Retorna uma lista de objetos Municipio
        com todos os municípios de um Estado.
        """""
        if self._munic_objs and not force:
            return self._munic_objs
        executor = concurrent.futures.ThreadPoolExecutor(max_workers=32)
        munics_ids = list(self._baixar_municipios().values())  # Pega os ids dos municípios
        munic = list(executor.map(
            lambda munic_id: Municipio(self._headers, self._id, munic_id), munics_ids
        ))  # Cria as instâncias
        self._munic_objs = [m for m in munic if len(m) > 0]
        return self._munic_objs  # retorna somente os municípios que tem mandados

    def _set_id(self, id_estado):
        if isinstance(id_estado, str):
            id_estado = id_estado.upper()
            if id_estado in UF_MAP:
                return UF_MAP[id_estado]
        elif isinstance(id_estado, int):
            if id_estado in NUM_MAP:
                return id_estado
        raise EstadoNotFoundError("id_estado precisa ser a sigla de uma UF ou um número entre 1 e 27")

    def _set_sigla(self):
        return NUM_MAP[self._id]

    def __len__(self):
        return self._totalElements

    def __add__(self, val):
        if not isinstance(val, (Estado, Municipio, OrgaoExpedidor)):
            raise TypeError("Você só pode somar elementos das classes Estado, Municipio ou OrgaoExpedidor")
        return self.obter_mandados() + val.obter_mandados()

    sigla = property(_set_sigla, None)


class Municipio(Filtro):
    def __init__(self, headers, id_estado: int, id_munic: int):
        super().__init__(headers)
        self._id_estado = id_estado
        self._id = id_munic
        self._init_info = self._obter_post_pag1(self._id_estado, self._id)
        self._totalElements = self._init_info['totalElements']
        self._mandados_list = []
        self._orgaos_info = {}
        self._orgaos_objs = []

    def baixar_mandados(self, force: bool = False):
        if self._mandados_list and not force:
            return self._mandados_list
        if self._totalElements <= 20000:
            if self._totalElements <= 10000:
                mandados_parciais = self._request_post_poucos_mandados(self._init_info, self._id_estado, self._id)
            else:
                mandados_parciais = self._request_post_expandido(self._init_info, self._id_estado, self._id)
            self._mandados_list = self._baixar_conteudo_completo_parallel(mandados_parciais)
        else:
            orgaos = self._gerar_orgaos()
            executor = concurrent.futures.ThreadPoolExecutor(max_workers=32)
            list(tqdm(executor.map(OrgaoExpedidor.baixar_mandados, orgaos), total=len(orgaos)))
            self._mandados_list = [item for orgao in orgaos for item in orgao.obter_mandados()]
        # print(f"Recuperados {len(self._mandados_list)}/{self._totalElements} mandados")

    def obter_mandados(self) -> list:
        if not self._mandados_list:
            raise MandadosNotFoundError("Você ainda não baixou os mandados. Utilize .baixar_mandados() para baixa-los")
        return self._mandados_list.copy()

    def _baixar_orgaos(self, force: bool = False) -> dict:
        """
        A partir do id da de um município retorna
        um dict mapeando todos os Órgãos Expeditores
        desse município aos seus ids.
        """
        if self._orgaos_info and not force:
            return self._orgaos_info
        response_org = requests.get(
            url=f'https://portalbnmp.cnj.jus.br/bnmpportal/api/pesquisa-pecas/orgaos/municipio/{self._id}',
            headers=HEADERS
        )
        org_list = response_org.json()
        return {org['nome']: org['id'] for org in org_list}

    def obter_orgaos(self) -> list:
        """
        Retorna uma lista de ids dos municipios dessa uf.
        """
        return list(self._baixar_orgaos().keys())

    def _gerar_orgaos(self, force: bool = False) -> list:
        """"
        Retorna uma lista de objetos Municipio
        com todos os municípios de um Estado.
        """""
        if self._orgaos_objs and not force:
            return self._orgaos_objs
        executor = concurrent.futures.ThreadPoolExecutor(max_workers=32)
        orgaos_ids = list(self._baixar_orgaos().values())  # Pega os ids dos órgãos expedidores
        org = list(executor.map(
            lambda org_id: OrgaoExpedidor(self._headers, self._id_estado, self._id, org_id), orgaos_ids
        ))  # Cria as instâncias
        return [o for o in org if len(o) > 0]  # retorna somente os orgaos que tem mandados

    def __len__(self):
        return self._totalElements

    def __add__(self, val):
        if not isinstance(val, (Estado, Municipio, OrgaoExpedidor)):
            raise TypeError("Você só pode somar elementos das classes Estado, Municipio ou OrgaoExpedidor")
        return self.obter_mandados() + val.obter_mandados()


class OrgaoExpedidor(Filtro):
    def __init__(self, headers, id_estado, id_munic, id_orgao):
        super().__init__(headers)
        self._id_estado = id_estado
        self._id_munic = id_munic
        self._id = id_orgao
        self._init_info = self._obter_post_pag1(self._id_estado, self._id_munic, self._id)
        self._totalElements = self._init_info['totalElements']
        self._mandados_list = []

    def baixar_mandados(self, force: bool = False):
        if self._mandados_list and not force:
            return self._mandados_list
        if self._totalElements <= 20000:
            if self._totalElements <= 10000:
                mandados_parciais = self._request_post_poucos_mandados(
                    self._init_info, self._id_estado, self._id_munic, self._id
                )
            else:
                mandados_parciais = self._request_post_expandido(
                    self._init_info, self._id_estado, self._id_munic, self._id
                )
            self._mandados_list = self._baixar_conteudo_completo_parallel(mandados_parciais)
        else:
            mandados_parciais = self._request_post_forcabruta(self._id_estado, self._id_munic, self._id)
            self._mandados_list = self._baixar_conteudo_completo_parallel(mandados_parciais)
        # print(f"Recuperados {len(self._mandados_list)}/{self._totalElements} mandados")

    def obter_mandados(self) -> list:
        if not self._mandados_list:
            raise MandadosNotFoundError("Você ainda não baixou os mandados. Utilize .baixar_mandados() para baixa-los")
        return self._mandados_list.copy()

    def __len__(self):
        return self._totalElements

    def __add__(self, val):
        if not isinstance(val, (Estado, Municipio, OrgaoExpedidor)):
            raise TypeError("Você só pode somar elementos das classes Estado, Municipio ou OrgaoExpedidor")
        return self.obter_mandados() + val.obter_mandados()
